package communication.bulletinboard;

import java.util.Iterator;
import java.util.concurrent.ConcurrentHashMap;

import communication.bulletinboard.messages.Message;

import exceptions.TagNotSupportedException;

/**
 * All the communication among nodes must be performed via the bulletin board.
 * The bulletin board MUST NOT support delete or edit operations (it is append-only)
 * and can be consulted by any node (or third party). If a confidential channel
 * between the nodes A and B is required, the messages from A MUST BE encrypted with
 * B's public key and viceversa. Likewise, an authenticated channel is achieved by 
 * attaching a digital signature to the message.
 * 
 * [Idea: AuthenticatedMessage and EncryptedMessage can be instances of Message to
 * enable this functionality. Currently, only plaintext messages are supported.]
 * @author SergioA
 *
 */
public class BulletinBoard {
	
	/**
	 * Singleton BulletinBoard instance
	 */
	private static BulletinBoard instance;
	
	/**
	 * The bulletin board is represented by a HashMap where the keys are tags
	 * defined in BBT
	 */
	private ConcurrentHashMap<BBT, BulletinSubject> bulletinBoard;
	
	private static final BBT[] SUPPORTED_TAGS = BBT.values();

	/**
	 * Initialize the bulletin board, registering the categories supported
	 */
	public BulletinBoard() 
	{
		bulletinBoard = new ConcurrentHashMap<>();
		for (int i = 0; i < SUPPORTED_TAGS.length; i++) {
			bulletinBoard.put(SUPPORTED_TAGS[i], new BulletinSubject());
		}
	}
	
	/**
	 * Returns the entry in the board with the subject marked by [tag]
	 * @param tag
	 * @return Messages under [tag]
	 * @throws TagNotSupportedException
	 */
	public BulletinSubject getSubject(BBT tag) throws TagNotSupportedException
	{
		if(!bulletinBoard.containsKey(tag))
			throw new TagNotSupportedException(tag);
		
		return bulletinBoard.get(tag);
	}
	
	/**
	 * Returns an iterator to all the messages under a specific tag
	 * @param tag
	 * @return Iterator of the messages under [tag]
	 * @throws TagNotSupportedException
	 */
	public Iterator<Message> getMessageIteratorByTag(BBT tag) throws TagNotSupportedException
	{
		BulletinSubject bulletinSubject = getSubject(tag);
		return bulletinSubject.getMessageIterator();
	}
	
	/**
	 * Register [newMessage] under [tag]
	 * @param tag
	 * @param newMessage
	 * @throws TagNotSupportedException
	 */
	public void addMessage(BBT tag, Message newMessage) throws TagNotSupportedException
	{
		BulletinSubject subject = getSubject(tag);
		subject.addMessage(newMessage);
	}
	
	public static BulletinBoard getInstance()
	{
		if(instance==null)
			instance = new BulletinBoard();
		return instance;
	}
	
	/**
	 * Output format:
	 * 
	 * "[Beginning of buffer]"
	 * 
	 * FOR EACH TAG <subject tag name>"\n"<string representation of the subject>"\n"
	 * 
	 * "[End of buffer]"
	 */
	public String toString()
	{
		StringBuilder bulletinContentBuffer = new StringBuilder();

		bulletinContentBuffer.append("[Beginning of buffer]\n\n");
		
		for (int i = 0; i < SUPPORTED_TAGS.length; i++) {					
			String subjectContent = bulletinBoard.get(SUPPORTED_TAGS[i]).toString();
			
			bulletinContentBuffer.append(SUPPORTED_TAGS[i]);
			bulletinContentBuffer.append("\n");
			bulletinContentBuffer.append(subjectContent);
		}
		bulletinContentBuffer.append("\n");
		
		bulletinContentBuffer.append("[End of buffer]");
		
		return bulletinContentBuffer.toString();
	}
}
