package communication.card;

/**
 * APDU commands: byte arrays that allow the communication with the smart card.
 * The general format of an APDU request is
 * CLA INS P1 P2 L [Data]
 * 
 * The first 5 bytes are mandatory. The content of data is optional and has a variable length.
 * 
 * Where:
 * CLA : Class identifier. Usually the instructions associated with a specific applet have
 * 		the same value of CLA.
 * INS : Instruction identifier.
 * P1, P2: Parameter bytes.
 * L: Length of data.
 * 
 * The SELECT APDU is standard. The rest are application specific. For a list of other standard 
 * APDUs, refer to http://web.archive.org/web/20090630004017/http://cheef.ru/docs/HowTo/APDU.info
 * 
 * The applet ID was arbitrarily fixed to 22 34 56 54 32 28 02
 * The CLA used for the instructions is 0x80
 * 
 * @author SergioA
 */
public enum APDUL {
	
	/**
	 *  Misc
	 */
	
	// Empty array
	EMPTY(new byte[]{}),
	
	// Requests to send more bytes from a previously started transmission
	SEND_MORE(new byte[]{0x00,(byte) 0xc0,0x00,0x00,0x00}),
	
	// 1-byte array containing the value 0x00
	ZERO_BYTE(new byte[]{0x00}),
	
	/**
	 * Selection and initialization
	 */
	
	// Select ElGamal applet: 00 A4 04 00 07 22 34 56 54 32 28 02
	SELECT_APPLET(new byte[]{0x00,(byte) 0xA4,0x04, 0x00, 0x07, 0x22, 0x34, 0x56, 0x54, 0x32, 0x28, 0x02}), // from CRISES.
	
	// Initialize the applet with 2048-bit ElGamal parameters
	INIT_2048(new byte[]{(byte) 0x80,0x00,0x08,0x00}),
	
	/**
	 *  Send parameters
	 */
	
	// Send value P
	SEND_P(new byte[]{0x01,0x00,0x00}),
	
	// Send value Q
	SEND_Q(new byte[]{0x11,0x00,0x00}),
	
	// Send value G
	SEND_G(new byte[]{0x02,0x00,0x00}),
	
	// Send threshold values t,n
	SEND_T_N(new byte[]{(byte) 0x80,0x1e}),
	
	// Prefix of APDU for sending the card ID
	SEND_CARD_ID_PRE(new byte[]{(byte) 0x80,0x12,0x00,0x00,0x01}),
	
	// Suffix of APDU for sending the card ID
	SEND_CARD_ID_POS(new byte[]{0x00}),
	
	/**
	 *  Trigger in-card operations
	 */
	
	// Requests the card to calculate a new private key
	PRIVATE_KEY_COMPUTE(new byte[]{(byte) 0x80,0x03,0x00,0x00}),
	
	// Requests the card to calculate a public key corresponding to the private key
	// already stored
	PUBLIC_KEY_COMPUTE(new byte[]{(byte) 0x80,0x04,0x00,0x00}),
	
	// Generates the coefficients for secret sharing
	GENERATE_COEFFICIENTS(new byte[]{(byte) 0x80,0x05,0x00,0x00}),
	
	// Generates the polynomial evaluation
	EVALUATE_POLYNOMIAL(new byte[]{(byte) 0x80,0x07,0x00,0x00}),
	
	// Generates the secret shares
	GENERATE_SHARES(new byte[]{(byte) 0x80,0x08,0x00,0x00}),
	
	/**
	 * Retrieve data from card
	 */
	
	// Get the public key
	GET_PK(new byte[]{(byte) 0x80,0x1A,0x00,0x00,0x00}),
	
	// Get share of the secret key
	GET_SK_SHARE(new byte[]{(byte) 0x80,0x19});
	

	public final byte[] apdu;
	
	APDUL(byte[] apdu)
	{
		this.apdu = apdu;
	}
}
