package secretsharing;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.math.RoundingMode;
import java.util.Random;

import encoding.utils.RandomnessUtils;
import threshold.elgamal.configuration.parameters.ThresholdElGamalPublicParameters;

/**
 * Enables the generation and merger of secret shares according to Shamir's
 * Secret Sharing scheme
 * 
 * @author SergioA
 *
 */
public class ShamirSecretSharing {

	/**
	 * Information about the threshold
	 */
	private ThresholdElGamalPublicParameters tegpp;

	/**
	 * PRNG instance
	 */
	private Random randS;

	public ShamirSecretSharing(ThresholdElGamalPublicParameters tegpp) {
		this.tegpp = tegpp;
		randS = RandomnessUtils.getRandomGenerator();
	}

	/**
	 * The parameters for the scheme are defined in the object [tegpp].
	 * Distributes a BigInteger secret in [n] shares such that [t] of them are
	 * required to retrieve the secret.
	 * 
	 * This implementation produces a polynomial of degree [t-1] where
	 * f(0)=[secret] and evaluates it for each integer j in (0,k-1).
	 * 
	 * @param [secret]
	 * @return [n] shares of [secret]
	 */
	public BigInteger[] generateShares(BigInteger secret) {
		BigInteger[] polynomial = new BigInteger[tegpp.getT()];

		// f(0)=secret
		polynomial[0] = secret;

		// Initialize the coefficients
		for (int i = 1; i < polynomial.length; i++) {
			polynomial[i] = new BigInteger(tegpp.getP().bitLength(), randS);
		}

		// Evaluate the polynomial with coefficients defined in [polynomial]
		BigInteger[] evaluations = new BigInteger[tegpp.getN()];
		for (int i = 0; i < evaluations.length; i++) {
			evaluations[i] = evaluatePolynomial(polynomial, i + 1);
		}

		return evaluations;
	}

	/**
	 * Given the coefficients of a polynomial, returns its evaluation in the
	 * point[i]. f(i) = coefficients[0] +
	 * coefficients[1]*i^1+...coefficients[k-1]*i^(k-1)
	 * 
	 * @param polynomial
	 * @param i
	 * @return
	 */
	private BigInteger evaluatePolynomial(BigInteger[] coefficients, int i) {
		// The first term of the polynomial is just the first coefficient
		BigInteger evaluation = coefficients[0];
		BigInteger bigIntegerI = BigInteger.valueOf(i);

		// Evaluation of each of the terms of the polynomial
		for (int j = 1; j < coefficients.length; j++) {
			BigInteger exponent = bigIntegerI.pow(j);
			BigInteger term = coefficients[j].multiply(exponent);
			evaluation = evaluation.add(term);
		}

		return evaluation;
	}

	/**
	 * Given a set of secret shares of the form [shamirShares]={s1,...,sn},
	 * generates an array of public commitments to those shares (i.e. public
	 * shares), such that [publicShares]={g^s1,...,g^sn}
	 * 
	 * @param shamirShares
	 * @return
	 */
	public BigInteger[] calculatePublicShares(BigInteger[] shamirShares) {
		BigInteger[] publicShares = new BigInteger[shamirShares.length];

		for (int i = 1; i < publicShares.length; i++) {
			publicShares[i] = tegpp.getG()
					.modPow(shamirShares[i], tegpp.getP());
		}

		return publicShares;
	}

	/**
	 * Given a set of secret shares expressed as points of the form (x,y),
	 * reconstructs the secret value.
	 * 
	 * A set of EXACTLY [t] points is necessary to perform a correct
	 * reconstruction.
	 * 
	 * @param points
	 * @return secret reconstructed
	 */
	public BigInteger reconstructSecret(BigInteger[][] points) {
		BigDecimal[] sum = new BigDecimal[2];
		sum[0] = BigDecimal.ZERO;
		sum[1] = BigDecimal.ONE;
		for (int i = 0; i < points.length; i++) {
			BigDecimal[] bd = singleLagrange(points, i);
			sum[0] = (sum[0].multiply(bd[1])).add(sum[1].multiply(bd[0]));
			sum[1] = sum[1].multiply(bd[1]);
		}
		BigInteger secret = sum[0].divide(sum[1], 0, RoundingMode.HALF_EVEN)
				.toBigInteger();
		return secret;
	}

	/**
	 * Calculates the lagrange multiplier required for the interpolation.
	 * 
	 * @param points
	 * @param i
	 *            : first index of the points evaluated
	 * @return lagrange multiplier
	 */
	public BigDecimal[] singleLagrange(BigInteger[][] points, int i) {
		BigDecimal[] multiplier = new BigDecimal[2];
		multiplier[0] = new BigDecimal("1");
		multiplier[1] = new BigDecimal("1");
		for (int k = 0; k < points.length; k++) {
			if (k != i) {
				BigDecimal a = BigDecimal.ZERO.subtract((new BigDecimal(
						points[k][0])));
				BigDecimal b = (new BigDecimal(points[i][0]))
						.subtract(new BigDecimal(points[k][0]));
				multiplier[0] = multiplier[0].multiply(a);
				multiplier[1] = multiplier[1].multiply(b);
			}
		}
		multiplier[0] = multiplier[0].multiply(new BigDecimal(points[i][1]));
		return multiplier;
	}

}
