package threshold.parties.brokers;

import java.math.BigInteger;
import java.security.NoSuchAlgorithmException;
import java.util.Iterator;
import java.util.Random;

import communication.bulletinboard.BBT;
import communication.bulletinboard.BulletinBoard;
import communication.bulletinboard.messages.Message;
import encoding.utils.RandomnessUtils;
import exceptions.PublicKeyNotFoundException;
import exceptions.TagNotSupportedException;
import threshold.elgamal.configuration.parameters.ElGamalPublicParameters;
import threshold.elgamal.entities.ElGamalCiphertext;
import threshold.elgamal.entities.ElGamalPublicKey;

/**
 * A threshold decryption scheme usually affects the key generation and the decryption,
 * but not the encryption. In particular in the schemes pursued, the encryption is assumed 
 * to be performed by a third party and won't involve smart cards. As a result,
 * this implementation handles it lightly, generating a simple encryption given the
 * public key parameters and the message.
 * @author SergioA
 *
 */
public class ElGamalOracle 
{
	/**
	 * Encrypts [message] using standard ElGamal and the parameters defined in 
	 * [params]
	 * @param message
	 * @param params
	 * @return ElGamalCiphertext encryption of [message]
	 * @throws PublicKeyNotFoundException
	 * @throws TagNotSupportedException
	 * @throws NoSuchAlgorithmException
	 */
	public static ElGamalCiphertext elGamalEncrypt(BigInteger message, ElGamalPublicParameters params) throws PublicKeyNotFoundException, TagNotSupportedException, NoSuchAlgorithmException
	{
		Iterator<Message> iterator = BulletinBoard.getInstance().getMessageIteratorByTag(BBT.PUBLIC_KEY_TAG);
		if(!iterator.hasNext())
			throw new PublicKeyNotFoundException();
		
		byte[] rawPublicKey = iterator.next().getMessage();
		
		ElGamalPublicKey publicKey = new ElGamalPublicKey(rawPublicKey, params);
		BigInteger p = params.getP();
		BigInteger g = params.getG();
		
		Random random = RandomnessUtils.getRandomGenerator();
		BigInteger blindingFactor = new BigInteger(p.bitLength(), random);
		BigInteger c1 = g.modPow(blindingFactor, p);
		
		BigInteger c2Factor = publicKey.getH().modPow(blindingFactor, p);
		BigInteger c2 = message.multiply(c2Factor);
		
		return new ElGamalCiphertext(c1, c2);
	}
	
	/**
	 * Decrypts[ciphertext] using [privateKey] and the parameters defined in 
	 * [params].
	 * @param ciphertext
	 * @param privateKey
	 * @param params
	 * @return The decrypted message.
	 */
	public static BigInteger elGamalDecrypt(ElGamalCiphertext ciphertext, BigInteger privateKey, ElGamalPublicParameters params)
	{
		BigInteger c1 = ciphertext.getC1();
		BigInteger c2 = ciphertext.getC2();
		BigInteger p = params.getP();
		
		BigInteger c1Res = c1.modPow(privateKey, p);
		
//		BigInteger c1Res = new BigInteger(1,r);
		BigInteger  m = c2.divide(c1Res).mod(p);
		return m;
	}
}
