package threshold.parties.brokers;

import java.math.BigInteger;
import java.util.Iterator;

import secretsharing.ShamirSecretSharing;
import threshold.elgamal.configuration.parameters.ThresholdElGamalPublicParameters;
import threshold.elgamal.configuration.shareholderinfo.ShareholderInfo;
import tuplespaces.FlagTupleSpace;
import communication.bulletinboard.BBT;
import communication.bulletinboard.BulletinBoard;
import communication.bulletinboard.messages.KeyShareMessage;
import communication.bulletinboard.messages.Message;
import communication.bulletinboard.messages.MockSecretMessage;
import communication.bulletinboard.messages.ShareholderInfoMessage;
import exceptions.TagNotSupportedException;

/**
 * This thread supports the interactive part of the key generation protocol communicating
 * to the Java Card implementation.
 * @author SergioA
 *
 */
public class ThreadThresholdElGamalJavaCardKeyGenerator extends Thread {
	
	private ThresholdElGamalPublicParameters parameters;
	private BigInteger privateKey;
	private ShareholderInfo shareholderInfo;
	private BulletinBoard bb;
	private FlagTupleSpace fts;
	private ThresholdElGamalJavaCardBroker broker;

	

	public ThreadThresholdElGamalJavaCardKeyGenerator(
			ThresholdElGamalPublicParameters parameters, 
			ShareholderInfo shareholderInfo, ThresholdElGamalJavaCardBroker broker) {
		super();
		this.parameters = parameters;
		this.shareholderInfo = shareholderInfo;
		this.broker = broker;
		
		bb = BulletinBoard.getInstance();
		fts = FlagTupleSpace.getInstance();
	}

	/**
	 * The thread generates its part of the share and notifies about the end of the 
	 * process by leaving a flag on the FlagTupleSpace (FTS).
	 * 
	 * NOTE: The call to the FTS works, but does not fulfill the contract entirely.
	 *   Review and avoid using custom named flags. 
	 */
	public void run() {
        try {
			generateShare();
			fts.put("SHAREGen", shareholderInfo.toString());
		} catch (TagNotSupportedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
        
    }
	
	private void generateShare() throws TagNotSupportedException {
		// The full protocol must generate and verify commitments before
		// releasing the
		// public key. The family of classes MockCommitment* enable this point
		// insecurely.
		// A secure commitment implementation should be implemented.

		ShamirSecretSharing sss = new ShamirSecretSharing(parameters);
		BigInteger[] shamirShares = getSharesFromCard();
		BigInteger[] publicShares = sss.calculatePublicShares(shamirShares);

		// Publish the public shares of the secret key.
		broadcast(BBT.PUBLIC_PARTIAL_KEY_SHARE_TAG, publicShares);
		
		// Secretly send share[j] to P[j] for each party j
		sendSecretShare(shamirShares);
		notifyShareSending();
		waitForShares();
		
		// THE CONSISTENCY OF THE SHARES SHOULD BE VERIFIED
		mixShares();
	}
	
	private BigInteger[] getSharesFromCard() {
		BigInteger[] shares = new BigInteger[parameters.getN()];
		
		for (int i=0;i<parameters.getN();i++){
			int id_receiver_card = i;
			byte[] shareRaw = broker.getShare((byte)id_receiver_card); 
			shares[i] = new BigInteger(shareRaw);
		}
		return shares;
	}

	public BigInteger getPrivateKey()
	{
		return privateKey;
	}

	/**
	 * Produces a secret share of the private key by mixing the shares sent by the
	 * other brokers. Stores the value of the secret key at the parent broker
	 * (i.e. the broker that initiated the thread).
	 * @throws TagNotSupportedException
	 */
	private void mixShares() throws TagNotSupportedException {
		privateKey = BigInteger.ZERO;
		synchronized (bb) {
			Iterator<Message> secretMessageIterator = bb.getMessageIteratorByTag(BBT.SECRET_PARTIAL_KEY_SHARE_TAG);
			while(secretMessageIterator.hasNext())
			{
				MockSecretMessage secretMessage = (MockSecretMessage) secretMessageIterator.next();
				if(secretMessage.getRecipientInfo().equals(shareholderInfo))
				{
					BigInteger shareValue = new BigInteger(secretMessage.getMessage());
					privateKey=privateKey.add(shareValue);
				}
			}
			broker.setPrivateKey(privateKey);
		}
	}

	/**
	 * The method verifies for each broker if the flag notifying the production
	 * of the share was published. If not, the thread blocks until the share is 
	 * produced. 
	 * @throws TagNotSupportedException
	 */
	private void waitForShares() throws TagNotSupportedException {
		Iterator<Message> shareholderInfoIterator = bb.getSubject(
				BBT.SHAREHOLDER_DATA_TAG).getMessageIterator();
		int i = 0;
		while (shareholderInfoIterator.hasNext()) {
			ShareholderInfoMessage simMessage = (ShareholderInfoMessage) shareholderInfoIterator
					.next();
			ShareholderInfo peerShareholderInfo = simMessage
					.getShareholderInfo();
			if (!peerShareholderInfo.equals(shareholderInfo)) {
				fts.read("SS", peerShareholderInfo.toString());
			}
			i++;
		}
	}

	private void notifyShareSending() {
		// TODO Auto-generated method stub
		fts.put("SS", shareholderInfo.toString());
	}

	/**
	 * Publishes a set of BigInteger values under the tag selected.
	 * @param tag
	 * @param setOfValues
	 * @throws TagNotSupportedException
	 */
	private void broadcast(BBT tag, BigInteger[] setOfValues) throws TagNotSupportedException {
		// TODO Auto-generated method stub
		for (int i = 1; i < setOfValues.length; i++) {
			KeyShareMessage secretShareMessage = new KeyShareMessage(
					shareholderInfo, i, setOfValues[i].toByteArray());
			bb.addMessage(tag,
					secretShareMessage);
		}
	}
	
	/**
	 * Sends secretly each share generated to the corresponding broker.
	 * NOTE: The current ShareholderInfo implementation, BasicShareholderInfo, 
	 * contains a WEAK implementation of this security promise.
	 * @param shamirShares
	 * @throws TagNotSupportedException
	 */
	private void sendSecretShare(BigInteger[] shamirShares) throws TagNotSupportedException
	{
		Iterator<Message> shareholderInfoIterator = bb.getSubject(
				BBT.SHAREHOLDER_DATA_TAG).getMessageIterator();
		int i = 0;
//		System.out.println(shareholderInfo.toString()+"> Sending "+shamirShares.length+" shares");
		while (shareholderInfoIterator.hasNext()) {
			ShareholderInfoMessage simMessage = (ShareholderInfoMessage) shareholderInfoIterator
					.next();
			ShareholderInfo recipientShareholderInfo = simMessage
					.getShareholderInfo();
//			if (!recipientShareholderInfo.equals(shareholderInfo)) {
				recipientShareholderInfo.sendMessage(
						BBT.SECRET_PARTIAL_KEY_SHARE_TAG,
						shamirShares[i].toByteArray());
//			}
			i++;
		}
	}
}
