package threshold.parties.brokers;

import java.math.BigInteger;
import java.util.Iterator;
import java.util.Random;

import communication.bulletinboard.BBT;
import communication.bulletinboard.BulletinBoard;
import communication.bulletinboard.BulletinSubject;
import communication.bulletinboard.messages.Message;
import communication.bulletinboard.messages.SimpleMessage;
import encoding.utils.RandomnessUtils;
import exceptions.TagNotSupportedException;
import threshold.elgamal.configuration.parameters.ThresholdElGamalPublicParameters;
import threshold.elgamal.configuration.shareholderinfo.ShareholderInfo;
import tuplespaces.FlagTupleSpace;

/**
 * This class uses components from Bouncy Castle's Diffie-Hellman key
 * generation. Although BouncyCastle is licensed for free usage, reusing its
 * code might be restricted. This usage is intended only for testing purposes
 * and should be revisited for future versions.
 * 
 * The original code is mostly taken from the class
 * org.bouncycastle.crypto.generators.DHKeyPairGenerator (and some helper
 * classes) The class can be found (alongwith the rest of the BouncyCastle code)
 * at http://grepcode.com/file/repo1.maven.org/maven2/bouncycastle/bcprov-jdk14/
 * 136/org/bouncycastle/crypto/generators/ElGamalKeyPairGenerator.java?av=f
 * 
 * @author SergioA
 *
 */

public class ThresholdElGamalJavaBroker implements Broker {

	private static final int MAX_ITERATIONS = 1000;
	
	private static final BigInteger ZERO = BigInteger.valueOf(0);
	private static final BigInteger TWO = BigInteger.valueOf(2);
	
	private BigInteger privateKey;

	private ThresholdElGamalPublicParameters parameters;

	private ShareholderInfo shareholderInfo;

	private ThresholdCryptographyEngine engine;
	
	private BulletinBoard bb;
	
	private FlagTupleSpace fts;

	private ThreadThresholdElGamalKeyGenerator keyGenerationThread;

	public ThresholdElGamalJavaBroker() {
		bb = BulletinBoard.getInstance();
		fts = FlagTupleSpace.getInstance();
	}

	@Override
	public void init(ThresholdElGamalPublicParameters parameters,
			ShareholderInfo shareholderInfo, ThresholdCryptographyEngine engine) {
		this.parameters = parameters;
		this.shareholderInfo = shareholderInfo;
		this.engine = engine;
	}

	private BigInteger calculatePrivate(BigInteger p, Random random,
			int limit) {
		//
		// calculate the private key
		//
		BigInteger pSub2 = p.subtract(TWO);
		BigInteger x;

		if (limit == 0) {
			x = createInRange(pSub2, random);
		} else {
			do {
				x = new BigInteger(limit, 0, random);
			} while (x.equals(ZERO));
		}

		return x;
	}

	private BigInteger createInRange(BigInteger max, Random random) {
		BigInteger x;
		int maxLength = max.bitLength();
		int count = 0;

		do {
			x = new BigInteger(maxLength, random);
			count++;
		} while ((x.equals(ZERO) || x.compareTo(max) > 0)
				&& count != MAX_ITERATIONS);

		if (count == MAX_ITERATIONS) // fall back to a faster (restricted)
										// method
		{
			return new BigInteger(maxLength - 1, random).setBit(0);
		}

		return x;
	}

	@Override
	public void generateShare() throws TagNotSupportedException {
		privateKey = calculatePrivate(parameters.getP(), RandomnessUtils.getRandomGenerator(), 0);
		
		// Public key share generation
		BigInteger publicKeyShare = parameters.getG().modPow(privateKey, parameters.getP());
		SimpleMessage publicKeyShareMessage = new SimpleMessage(publicKeyShare.toByteArray());
//		System.out.println("Key pair:"+shareholderInfo+":"+privateKey+":"+publicKeyShare);
		bb.addMessage(BBT.PUBLIC_KEY_SHARE_TAG, publicKeyShareMessage);
		
		keyGenerationThread = new ThreadThresholdElGamalKeyGenerator(parameters, privateKey, shareholderInfo, this);
		keyGenerationThread.start();
	}

//	@Override
//	public void shareDecryption(ElGamalCiphertext ciphertext) throws TagNotSupportedException 
//	{
////		BigInteger c1Res = new BigInteger(1,r);
//		BigInteger  m = ElGamalOracle.elGamalDecrypt(ciphertext, privateKey, parameters);
//		
//		bb.addMessage(bb.SHARE_DECRYPTION_TAG, new SimpleMessage(m.toByteArray()));
//	}

	@Override
	public void generatePublicKey() throws TagNotSupportedException {
		// TODO Auto-generated method stub
		BulletinSubject publicKeyShareBulletin = bb.getSubject(BBT.PUBLIC_KEY_SHARE_TAG);
		Iterator<Message> publicKeyShares = publicKeyShareBulletin.getMessageIterator();
		BigInteger publicKey= BigInteger.ONE;
		while(publicKeyShares.hasNext())
		{
			BigInteger publicKeyShare = new BigInteger(publicKeyShares.next().getMessage());
			publicKey= publicKey.multiply(publicKeyShare);
		}
		
		publicKey = publicKey.mod(parameters.getP());
		SimpleMessage publicKeyMessage = new SimpleMessage(publicKey.toByteArray());
		bb.addMessage(BBT.PUBLIC_KEY_TAG, publicKeyMessage);
		
	}
	
	protected void setPrivateKey(BigInteger privateKey) throws TagNotSupportedException
	{
		this.privateKey = privateKey;
//		System.out.println("MY PRIVATE KEY:"+privateKey);		
	}

	@Override
	public BigInteger getPrivateKeyShadow() throws TagNotSupportedException {
		// TODO Auto-generated method stub
		return privateKey;
	}
	
	

}
