package threshold.parties.brokers;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.math.BigInteger;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Random;

import javax.smartcardio.*;

import communication.bulletinboard.BBT;
import communication.bulletinboard.BulletinBoard;
import communication.bulletinboard.BulletinSubject;
import communication.bulletinboard.messages.Message;
import communication.bulletinboard.messages.SimpleMessage;
import communication.card.APDUL;
import encoding.utils.RandomnessUtils;
import encoding.utils.Utils;
import exceptions.TagNotSupportedException;
import threshold.elgamal.configuration.parameters.ThresholdElGamalPublicParameters;
import threshold.elgamal.configuration.shareholderinfo.ShareholderInfo;
import tuplespaces.FlagTupleSpace;

/**
 * This class uses components from Bouncy Castle's Diffie-Hellman key
 * generation. Although BouncyCastle is licensed for free usage, reusing its
 * code might be restricted. This usage is intended only for testing purposes
 * and should be revisited for future versions.
 * 
 * The original code is mostly taken from the class
 * org.bouncycastle.crypto.generators.DHKeyPairGenerator (and some helper
 * classes) The class can be found (alongwith the rest of the BouncyCastle code)
 * at http://grepcode.com/file/repo1.maven.org/maven2/bouncycastle/bcprov-jdk14/
 * 136/org/bouncycastle/crypto/generators/ElGamalKeyPairGenerator.java?av=f
 * 
 * Additionally, this class builds upon the implementation of the CRISES group
 * of a JavaCard-based threshold system. Some methods are directly taken from
 * their publicly available codebase (see
 * https://github.com/CRISES-URV/eVerification-2).
 * 
 * Further refinement of this class should take all the JavaCard communication
 * methods into a helper class, since they tend to be very verbose and
 * compromise the readability of the code.
 * 
 * @author SergioA
 *
 */

public class ThresholdElGamalJavaCardBroker implements Broker {

	private static final int MAX_ITERATIONS = 1000;

	private static final BigInteger ZERO = BigInteger.valueOf(0);
	private static final BigInteger TWO = BigInteger.valueOf(2);

//	private BigInteger privateKey;

	private ThresholdElGamalPublicParameters parameters;

	private ShareholderInfo shareholderInfo;

	private ThresholdCryptographyEngine engine;

	private BulletinBoard bb;

	private FlagTupleSpace fts;

	private ThreadThresholdElGamalJavaCardKeyGenerator keyGenerationThread;

	private CardTerminal terminal;

	private Card card;

	private CardChannel channel;

	private String[] responsesw;

	private BigInteger privateKey;

	public ThresholdElGamalJavaCardBroker() {
		bb = BulletinBoard.getInstance();
		fts = FlagTupleSpace.getInstance();
	}

	@Override
	public void init(ThresholdElGamalPublicParameters parameters,
			ShareholderInfo shareholderInfo, ThresholdCryptographyEngine engine) {
		this.parameters = parameters;
		this.shareholderInfo = shareholderInfo;
		this.engine = engine;

		TerminalFactory factory = TerminalFactory.getDefault(); // uses SunPCSC;
																// the unique
																// available to
																// interact with
																// PC/SC stack.
		try {
			List<CardTerminal> terminals = factory.terminals().list();
			terminal = terminals.get(shareholderInfo.getID());

			card = terminal.connect("T=1");
			System.out.print(" card: " + card);
			channel = card.getBasicChannel();

			// Get ATR (can be used to fingerprint the smart card).
			byte[] baATR = card.getATR().getBytes();
			System.out.println("ATR: " + Utils.toHexString(baATR));
			
			initializeCard();
		} catch (CardException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	/**
	 * Send [parameters] to the card
	 */
	private void initializeCard() {
		responsesw = new String[2];
		try {
			// Select Applet
			sendAPDU(APDUL.SELECT_APPLET.apdu);
			
			// Initialize with 2048-bit key length
			sendAPDU(APDUL.INIT_2048.apdu);
			
			// Send P
			byte[][] pAPDU = prepareAPDUSet(APDUL.SEND_P.apdu, parameters.getP()
					.toByteArray(), APDUL.EMPTY.apdu);
			sendAPDUSet(pAPDU);
			
			// Send Q
			 byte[][] qAPDU = prepareAPDUSet(APDUL.SEND_Q.apdu,
			 parameters.getQ().toByteArray(), APDUL.EMPTY.apdu);
			 sendAPDUSet(qAPDU);

			// Send G
			byte[][] gAPDU = prepareAPDUSet(APDUL.SEND_G.apdu, parameters.getG()
					.toByteArray(), APDUL.EMPTY.apdu);
			sendAPDUSet(gAPDU);

			// Send T,N
			byte[] a = APDUL.SEND_T_N.apdu;
			byte[] tnAPDU = new byte[]{a[0], a[1],(byte)parameters.getT(), (byte) parameters.getN()};
			sendAPDU(tnAPDU);

			// //send id_card
			// send( "80120000010"+id_card+"00",
			// "Save id card = 0 (it must be in 0..n)", id_card);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	private byte[] trimTrailingZeros(byte[] byteArray) 
	{
		int trail = 0;
		for (int i = 0; i < byteArray.length && byteArray[i]==0x00; i++) {
			trail++;
		}
		byte[] trimmedArray = new byte[byteArray.length-trail];
		System.arraycopy(byteArray, trail, trimmedArray, 0, trimmedArray.length);
		return trimmedArray;
		
	}

	/**
	 * Prepares a set of APDUs based on the [apduPrefix], [content] and [suffix]
	 * provided. It is designed to support the transmission of long [content] elements.
	 * The maximum [content] length per APDU is 32 bytes. 
	 * @param apduPrefix
	 * @param content
	 * @param suffix
	 * @return
	 */
	private byte[][] prepareAPDUSet(byte[] apduPrefix, byte[] content, byte[] suffix) {
		byte CLA = (byte) 0x00;
		byte LC = (byte) 0x20;
		byte[] subdata ;
		ArrayList<byte[]> apdus = new ArrayList<>();
		
		content = trimTrailingZeros(content);
		
		int index = content.length;
		
		while(index > 0)
		{
			if(index > 32)
			{
				CLA = (byte) 0x90;
			}
			else
			{
				CLA = (byte) 0x80;
			}
			
			// subdata size = prefix_size + data[index-32:index] + suffix_size
			LC  = 0x20;

			byte[] singleAPDU = prepareAPDU(CLA, LC, apduPrefix, content, index);
			
//			subdata[subdata.length-1] = 0x00;
			
			apdus.add(singleAPDU);
			
			if(index>32)
			{
				content = Arrays.copyOf(content, index-32);
				index = content.length;
			}
			else
			{
				index=0;
			}
		}
		
		// ArrayList parsing
		byte[][] apdusArray = new byte[apdus.size()][];
		apdus.toArray(apdusArray);
		return apdusArray;
	}
	
	/**
	 * Prepares a partial APDU with the bytes from content[index-32] to 
	 * content[index]
	 * @param CLA
	 * @param LC
	 * @param apduPrefix
	 * @param content
	 * @param index
	 * @return
	 */
	private byte[] prepareAPDU(byte CLA, byte LC, byte[] apduPrefix, byte[] content, int index)
	{
		byte[] apduArray = new byte[5+ 32 + 1 ];
		content = padLeftByteArray(content, 32);
		apduArray = setPrefix(apduPrefix, apduArray, CLA, LC);
		
		if(index< 32)index=32;
		// Copy the next 32 bytes from [content] to [subdata] 
		System.arraycopy(content, index-32, apduArray, 5, 32);
		
		return apduArray;
	}
	
	/**
	 * Creates an array based on [array]. If [size] is longer than the size of
	 * [array], the new array is left-padded with the byte 0x00 until the length
	 * matches.
	 * @param array
	 * @param size
	 * @return leftpadded(array)
	 */
	private byte[] padLeftByteArray(byte[] array, int size)
	{
		if(array.length < size)
		{
			byte[] temp = new byte[size];
			int paddingSize = size - array.length;
			System.arraycopy(array, 0, temp, paddingSize, array.length);
			return temp;
		}
		return array;
	}
	
	/**
	 * Copies the header information of an APDU to [targetArray]. If the [prefix]
	 * is shorter than the size of the header, the header is right-padded with 
	 * 0x00 bytes.
	 * 
	 * An APDU header has the form:
	 * - INS CLA P1 P2 LC
	 * 
	 * [prefix] may have the forms:
	 * - CLA
	 * - CLA P1
	 * - CLA P1 P2
	 * @param prefix
	 * @param targetArray
	 * @param CLA
	 * @param LC
	 * @return [targetArray]
	 */
	private byte[] setPrefix(byte[] prefix, byte[] targetArray, byte INS, byte LC)
	{
		targetArray[0] = INS;
		
		// Fill the header with the [apduPrefix] data, or null bytes if the prefix is too short
		for(int i = 1; i < 4 ; i++)
		{
			if(i < prefix.length)
			{
				targetArray[i]=prefix[i-1];
			}
			else
			{
				targetArray[i]=0x00;
			}
			
		}
		
		// 5th byte = LC
		targetArray[4] = LC;
		
		return targetArray;
	}

	/**
	 * Generates the part of the secret corresponding to this broker. Initiates the
	 * thread that performs the interactive part to agree upon a secret.
	 */
	@Override
	public void generateShare() throws TagNotSupportedException {

		try {
			// Send private key initialization instruction
			sendAPDU(APDUL.PRIVATE_KEY_COMPUTE.apdu);

			// Send public key initialization instruction
			sendAPDU(APDUL.PUBLIC_KEY_COMPUTE.apdu);
			

			sendAPDU(APDUL.GENERATE_COEFFICIENTS.apdu);
			// TODO HERE SHOULD GENERATE COMMITMENTS TO THE COEFFICIENTS
			sendAPDU(APDUL.EVALUATE_POLYNOMIAL.apdu);
			
			sendAPDU(APDUL.GENERATE_SHARES.apdu);
			// TODO HERE SHOULD GENERATE COMMITMENTS TO THE SHARES
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		// Public key share generation
		BigInteger publicKeyShare = getPublicKey();
		
		SimpleMessage publicKeyShareMessage = new SimpleMessage(
				publicKeyShare.toByteArray());
		// System.out.println("Key pair:"+shareholderInfo+":"+privateKey+":"+publicKeyShare);
		bb.addMessage(BBT.PUBLIC_KEY_SHARE_TAG, publicKeyShareMessage);

		keyGenerationThread = new ThreadThresholdElGamalJavaCardKeyGenerator(
				parameters, shareholderInfo, this);
		keyGenerationThread.start();
	}
	
	// @Override
	// public void shareDecryption(ElGamalCiphertext ciphertext) throws
	// TagNotSupportedException
	// {
	// // BigInteger c1Res = new BigInteger(1,r);
	// BigInteger m = ElGamalOracle.elGamalDecrypt(ciphertext, privateKey,
	// parameters);
	//
	// bb.addMessage(bb.SHARE_DECRYPTION_TAG, new
	// SimpleMessage(m.toByteArray()));
	// }

	private BigInteger getPublicKey() {
		try {
			byte[] pkArray = sendAPDU(APDUL.GET_PK.apdu);
			BigInteger pk = new BigInteger(pkArray);
			return pk;
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return null;
	}

	@Override
	public void generatePublicKey() throws TagNotSupportedException {
		// TODO Auto-generated method stub
		BulletinSubject publicKeyShareBulletin = bb
				.getSubject(BBT.PUBLIC_KEY_SHARE_TAG);
		Iterator<Message> publicKeyShares = publicKeyShareBulletin
				.getMessageIterator();
		BigInteger publicKey = BigInteger.ONE;
		while (publicKeyShares.hasNext()) {
			BigInteger publicKeyShare = new BigInteger(publicKeyShares.next()
					.getMessage());
			publicKey = publicKey.multiply(publicKeyShare);
		}

		publicKey = publicKey.mod(parameters.getP());
		SimpleMessage publicKeyMessage = new SimpleMessage(
				publicKey.toByteArray());
		bb.addMessage(BBT.PUBLIC_KEY_TAG, publicKeyMessage);

	}
	
	public byte[] getShare(byte recipientID)
	{
		try {
			byte[] shareAPDU = prepareShortAPDU(APDUL.GET_SK_SHARE.apdu, new byte[]{recipientID}, APDUL.ZERO_BYTE.apdu, APDUL.ZERO_BYTE.apdu);
			byte[] r = sendAPDU(shareAPDU);
			return r;
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
//		send( "80190"+id_receiver_card+"00", "Get share["+id_receiver_card+"] from N shares", shareholderInfo.getID());
		return null;
	}

	private byte[] prepareShortAPDU(byte[]... parts) {
		// TODO Auto-generated method stub
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		for (int i = 0; i < parts.length; i++) {
			try {
				baos.write(parts[i]);
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		return baos.toByteArray();
	}

	/**
	 * [Method directly from the CRISES implementation] Used to send the
	 * information to a SC. In this case the APDU is fractioned in the following
	 * parameters.
	 * 
	 * @param apdu
	 *            The apdu in a only block
	 * @param text
	 *            The text
	 * @param id_card
	 *            The identifier of the SC
	 * @return The return message of the SC
	 * @throws CardException
	 */
	private String send(String apdu, String text) throws CardException {
		CommandAPDU c;
		String ret = "0";
		String sw = "";

		if (!apdu.isEmpty()) {

			// if (currentsizeinbits == 2048 && apdu.length() == 8 ){ //hacking
			// per 2048 bits -> java.lang.ArrayIndexOutOfBoundsException: 4
			// apdu += "00";
			// }
			c = new CommandAPDU(Utils.hexStringToByteArray(apdu));

			ResponseAPDU r = channel.transmit(c);
			// System.out.print("<--"+id_card+"--- ");
			sw = Integer.toHexString(r.getSW());
			responsesw[0] = sw;
			responsesw[1] = String.valueOf(shareholderInfo.getID());

			// System.out.println("ret(SW): "+ sw);
			if (r.getData().length > 0) {
				System.out.println(r.getData());
				ret = Utils.toHexString(r.getData());

				while (sw.compareTo("ffc0") == 0) { // ISO7816.SW_BYTES_REMAINING_00
					// System.out.println("---"+id_card+"--> " + "00c0000000");
					c = new CommandAPDU(
							Utils.hexStringToByteArray("00c0000000")); // send
																		// more
					r = channel.transmit(c);
					ret = Utils.toHexString(r.getData()).concat(ret);
					sw = Integer.toHexString(r.getSW());
					this.responsesw[0] = sw;
					this.responsesw[1] = String
							.valueOf(shareholderInfo.getID());
					// System.out.print("<--"+id_card+"--- ");
					// System.out.println("ret(SW): "+sw);
				}
				// System.out.println("        Data returned: " + ret);
			} else {
				System.out.println();
			}
			System.out.println();
		} else {
			this.responsesw[0] = sw;
			this.responsesw[1] = String.valueOf(shareholderInfo.getID());
			System.out
					.println(" --------------------- WARNING: apdu is empty --------------------- ");
		}

		return ret;
	}
	
	private byte[][] sendAPDUSet(byte[][] apduSet) throws Exception
	{
		byte[][] responses = new byte[apduSet.length][];
		for (int i = 0; i < apduSet.length; i++) {
				responses[i] = sendAPDU(apduSet[i]);
		}
		
		return responses;
	}

	private byte[] sendAPDU(byte[] apdu) throws Exception {
		try {
			System.out.println("Sending:"+Utils.toHexString(apdu));
			CommandAPDU commAPDU = new CommandAPDU(apdu);
			ResponseAPDU r = channel.transmit(commAPDU);
			String sw = Integer.toHexString(r.getSW());
			ArrayList<byte[]> chunks = new ArrayList<>();
			
			
			responsesw[0]=sw;
			responsesw[1]=String.valueOf(shareholderInfo.getID());
			
			System.out.println("ret(SW): "+ sw);
			if (r.getData().length > 0){
//				System.out.println(r.getData());
				chunks.add(r.getData());
				
				while(sw.compareTo("ffc0")==0){ //ISO7816.SW_BYTES_REMAINING_00
					CommandAPDU c = new CommandAPDU(APDUL.SEND_MORE.apdu); //send more
					r = channel.transmit(c);
					chunks.add(r.getData());
					sw = Integer.toHexString(r.getSW());
					this.responsesw[0]=sw;
					this.responsesw[1]=String.valueOf(shareholderInfo.getID());
				}
			}else{
				System.out.println();
			}
			System.out.println();
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			for (int i = chunks.size()-1; i >=0; i--) {
				baos.write(chunks.get(i));
				
			}
			
			return baos.toByteArray();
		} catch (CardException e) {
			// TODO Auto-generated catch block
			throw new Exception("Error while sending the APDU "
					+ Utils.toHexString(apdu));
		}
	}

	@Override
	public BigInteger getPrivateKeyShadow() throws TagNotSupportedException {
		// TODO Auto-generated method stub
		return privateKey;
	}

	public void setPrivateKey(BigInteger privateKey) {
		// TODO Auto-generated method stub
		this.privateKey=privateKey;
	}

}
